from typing import Dict, Callable, Iterable, Tuple

import tensorflow as tf

from headliner.model.summarizer import Summarizer
from headliner.utils.logger import get_logger


class EvaluationCallback(tf.keras.callbacks.Callback):
    """
    Callback for custom scoring methods.
    """

    def __init__(self,
                 summarizer: Summarizer,
                 scorers: Dict[str, Callable[[Dict], float]],
                 val_data: Iterable[Tuple[str, str]],
                 print_num_examples=5) -> None:
        """
        Initializes the Callback.

        Args:
            summarizer: Summarizer that predicts over the validation data.
            scorers: Dictionary of {scorer_name: scorer}, where each scorer maps a prediction to a score.
            val_data: Raw validation data to predict on.
            print_num_examples: Number of prediction examples to output for eyeballing the prediction quality.
        """

        super().__init__()
        self.summarizer = summarizer
        self.scorers = scorers
        self.val_data = val_data
        self.logger = get_logger(__name__)
        self.print_num_examples = print_num_examples

    def on_epoch_end(self, batch, logs=None) -> None:
        if logs is None:
            logs = {}
        val_scores = {score_name: 0. for score_name in self.scorers.keys()}
        count_val = 0
        for d in self.val_data:
            count_val += 1
            input_text, target_text = d
            prediction = self.summarizer.predict_vectors(input_text, target_text)
            if count_val <= self.print_num_examples:
                self.logger.info('\n(input) {} \n(target) {} \n(prediction) {}\n'.format(
                    prediction['preprocessed_text'][0],
                    prediction['preprocessed_text'][1],
                    prediction['predicted_text']
                ))
            elif len(self.scorers) == 0:
                break
            for score_name, scorer in self.scorers.items():
                score = scorer(prediction)
                val_scores[score_name] += score
        for score_name, score in val_scores.items():
            logs[score_name] = float(score / count_val)



from typing import Tuple

from headliner.model.transformer_util import *


class Encoder(tf.keras.layers.Layer):

    def __init__(self,
                 num_layers: int,
                 num_heads: int,
                 feed_forward_dim: int,
                 embedding_shape: Tuple[int, int],
                 embedding_trainable=True,
                 embedding_weights=None,
                 dropout_rate=0.1,
                 max_seq_len=10000) -> None:
        super(Encoder, self).__init__()

        self.num_layers = num_layers
        vocab_size, vec_dim = embedding_shape
        weights = None if embedding_weights is None else [embedding_weights]
        self.embedding_size = vec_dim
        self.embedding = tf.keras.layers.Embedding(vocab_size,
                                                   vec_dim,
                                                   weights=weights,
                                                   trainable=embedding_trainable)
        self.pos_encoding = positional_encoding(max_seq_len, self.embedding_size)
        self.enc_layers = [EncoderLayer(vec_dim, num_heads, feed_forward_dim, dropout_rate)
                           for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(dropout_rate)

    def call(self, x, training, mask):
        seq_len = tf.shape(x)[1]
        x = self.embedding(x)
        x *= tf.math.sqrt(tf.cast(self.embedding_size, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]
        x = self.dropout(x, training=training)
        for i in range(self.num_layers):
            x = self.enc_layers[i](x, training, mask)
        return x


class Decoder(tf.keras.layers.Layer):

    def __init__(self,
                 num_layers: int,
                 num_heads: int,
                 feed_forward_dim: int,
                 embedding_shape: Tuple[int, int],
                 embedding_trainable=True,
                 embedding_weights=None,
                 dropout_rate=0.1,
                 max_seq_len=10000) -> None:
        super(Decoder, self).__init__()

        self.num_layers = num_layers
        vocab_size, vec_dim = embedding_shape
        weights = None if embedding_weights is None else [embedding_weights]
        self.embedding_size = vec_dim
        self.embedding = tf.keras.layers.Embedding(vocab_size,
                                                   vec_dim,
                                                   weights=weights,
                                                   trainable=embedding_trainable)
        self.pos_encoding = positional_encoding(max_seq_len, vec_dim)
        self.dec_layers = [DecoderLayer(vec_dim, num_heads, feed_forward_dim, dropout_rate)
                           for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(dropout_rate)

    def call(self,
             x,
             enc_output,
             training,
             look_ahead_mask,
             padding_mask):
        seq_len = tf.shape(x)[1]
        attention_weights = {}

        x = self.embedding(x)
        x *= tf.math.sqrt(tf.cast(self.embedding_size, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]
        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x, block1, block2 = self.dec_layers[i](x, enc_output, training,
                                                   look_ahead_mask, padding_mask)
            attention_weights['decoder_layer{}_block1'.format(i + 1)] = block1
            attention_weights['decoder_layer{}_block2'.format(i + 1)] = block2

        return x, attention_weights


class Transformer(tf.keras.Model):

    def __init__(self,
                 num_layers: int,
                 num_heads: int,
                 feed_forward_dim: int,
                 embedding_shape_encoder: Tuple[int, int],
                 embedding_shape_decoder: Tuple[int, int],
                 embedding_encoder_trainable=True,
                 embedding_decoder_trainable=True,
                 embedding_weights_encoder=None,
                 embedding_weights_decoder=None,
                 dropout_rate=0.1,
                 max_sequence_len=10000) -> None:
        super(Transformer, self).__init__()

        self.encoder = Encoder(num_layers,
                               num_heads,
                               feed_forward_dim,
                               embedding_shape_encoder,
                               embedding_trainable=embedding_encoder_trainable,
                               embedding_weights=embedding_weights_encoder,
                               dropout_rate=dropout_rate,
                               max_seq_len=max_sequence_len)

        self.decoder = Decoder(num_layers,
                               num_heads,
                               feed_forward_dim,
                               embedding_shape_decoder,
                               embedding_trainable=embedding_decoder_trainable,
                               embedding_weights=embedding_weights_decoder,
                               dropout_rate=dropout_rate,
                               max_seq_len=max_sequence_len)

        self.final_layer = tf.keras.layers.Dense(embedding_shape_decoder[0])

    def call(self, inp, tar, training, enc_padding_mask,
             look_ahead_mask, dec_padding_mask):
        enc_output = self.encoder(inp, training, enc_padding_mask)

        dec_output, attention_weights = self.decoder(
            tar, enc_output, training, look_ahead_mask, dec_padding_mask)
        final_output = self.final_layer(dec_output)  # (batch_size, tar_seq_len, target_vocab_size)
        return final_output, attention_weights



import os
import pickle
from typing import Callable, Dict, Union

import numpy as np
import tensorflow as tf

from headliner.model.model_attention import Encoder, Decoder
from headliner.model.summarizer import Summarizer
from headliner.preprocessing.preprocessor import Preprocessor
from headliner.preprocessing.vectorizer import Vectorizer


class SummarizerAttention(Summarizer):

    def __init__(self, lstm_size=50, max_prediction_len=20, embedding_size=50, embedding_encoder_trainable=True,
                 embedding_decoder_trainable=True):
        super().__init__()
        self.lstm_size = lstm_size
        self.max_prediction_len = max_prediction_len
        self.embedding_size = embedding_size
        self.embedding_encoder_trainable = embedding_encoder_trainable
        self.embedding_decoder_trainable = embedding_decoder_trainable
        self.optimizer = SummarizerAttention._new_optimizer()
        self.encoder = None
        self.decoder = None
        self.embedding_shape_in = None
        self.embedding_shape_out = None

    def init_model(self,
                   preprocessor: Preprocessor,
                   vectorizer: Vectorizer,
                   embedding_weights_encoder=None,
                   embedding_weights_decoder=None) -> None:
        self.preprocessor = preprocessor
        self.vectorizer = vectorizer
        self.embedding_shape_in = (self.vectorizer.encoding_dim, self.embedding_size)
        self.embedding_shape_out = (self.vectorizer.decoding_dim, self.embedding_size)
        self.encoder = Encoder(self.embedding_shape_in,
                               self.lstm_size,
                               embedding_trainable=self.embedding_encoder_trainable,
                               embedding_weights=embedding_weights_encoder)
        self.decoder = Decoder(self.embedding_shape_out,
                               self.lstm_size,
                               embedding_trainable=self.embedding_decoder_trainable,
                               embedding_weights=embedding_weights_decoder)
        self.encoder.compile(optimizer=self.optimizer)
        self.decoder.compile(optimizer=self.optimizer)

    def __getstate__(self):
        """ Prevents pickle from serializing encoder and decoder """
        state = self.__dict__.copy()
        del state['encoder']
        del state['decoder']
        del state['optimizer']
        return state

    def predict(self, text: str) -> str:
        return self.predict_vectors(text, '')['predicted_text']

    def predict_vectors(self, input_text: str, target_text: str) -> Dict[str, Union[str, np.array]]:
        text_preprocessed = self.preprocessor((input_text, target_text))
        en_inputs, _ = self.vectorizer(text_preprocessed)
        en_initial_states = self.encoder.init_states(1)
        en_outputs = self.encoder(tf.constant([en_inputs]), en_initial_states)
        start_end_seq = self.vectorizer.encode_output(
            ' '.join([self.preprocessor.start_token, self.preprocessor.end_token]))
        de_start_index, de_end_index = start_end_seq[:1], start_end_seq[-1:]
        de_input = tf.constant([de_start_index])
        de_state_h, de_state_c = en_outputs[1:]
        output = {'preprocessed_text': text_preprocessed,
                  'logits': [],
                  'alignment': [],
                  'predicted_sequence': []}
        for _ in range(self.max_prediction_len):
            de_output, de_state_h, de_state_c, alignment = self.decoder(de_input, (de_state_h, de_state_c),
                                                                        en_outputs[0])
            de_input = tf.expand_dims(tf.argmax(de_output, -1), 0)
            pred_token_index = de_input.numpy()[0][0]
            if pred_token_index != 0:
                output['logits'].append(np.squeeze(de_output.numpy()))
                output['alignment'].append(np.squeeze(alignment.numpy()))
                output['predicted_sequence'].append(pred_token_index)
                if pred_token_index == de_end_index:
                    break
        output['predicted_text'] = self.vectorizer.decode_output(output['predicted_sequence'])
        return output

    def new_train_step(self,
                       loss_function: Callable[[tf.Tensor], tf.Tensor],
                       batch_size: int,
                       apply_gradients=True) -> Callable[[tf.Tensor, tf.Tensor], float]:

        train_step_signature = [
            tf.TensorSpec(shape=(batch_size, None), dtype=tf.int32),
            tf.TensorSpec(shape=(batch_size, self.vectorizer.max_output_len), dtype=tf.int32),
        ]
        encoder = self.encoder
        decoder = self.decoder
        optimizer = self.optimizer

        def train_step(source_seq, target_seq):
            loss = 0
            en_initial_states = encoder.init_states(batch_size)
            with tf.GradientTape() as tape:
                en_outputs = encoder(source_seq, en_initial_states)
                en_states = en_outputs[1:]
                de_state_h, de_state_c = en_states
                for i in range(target_seq.shape[1] - 1):
                    decoder_in = tf.expand_dims(target_seq[:, i], 1)
                    logit, de_state_h, de_state_c, _ = decoder(
                        decoder_in, (de_state_h, de_state_c), en_outputs[0])
                    loss += loss_function(target_seq[:, i + 1], logit)
            if apply_gradients is True:
                variables = encoder.trainable_variables + decoder.trainable_variables
                gradients = tape.gradient(loss, variables)
                optimizer.apply_gradients(zip(gradients, variables))
            return loss / (target_seq.shape[1] - 1)

        if self.vectorizer.max_output_len is not None:
            return tf.function(train_step, input_signature=train_step_signature)
        else:
            return train_step

    def save(self, out_path: str) -> None:
        if not os.path.exists(out_path):
            os.mkdir(out_path)
        summarizer_path = os.path.join(out_path, 'summarizer.pkl')
        encoder_path = os.path.join(out_path, 'encoder')
        decoder_path = os.path.join(out_path, 'decoder')
        with open(summarizer_path, 'wb+') as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)
        self.encoder.save_weights(encoder_path, save_format='tf')
        self.decoder.save_weights(decoder_path, save_format='tf')

    @staticmethod
    def load(in_path: str):
        summarizer_path = os.path.join(in_path, 'summarizer.pkl')
        encoder_path = os.path.join(in_path, 'encoder')
        decoder_path = os.path.join(in_path, 'decoder')
        with open(summarizer_path, 'rb') as handle:
            summarizer = pickle.load(handle)
        summarizer.encoder = Encoder(summarizer.embedding_shape_in,
                                     summarizer.lstm_size,
                                     embedding_trainable=summarizer.embedding_encoder_trainable)
        summarizer.decoder = Decoder(summarizer.embedding_shape_out,
                                     summarizer.lstm_size,
                                     embedding_trainable=summarizer.embedding_decoder_trainable)
        optimizer = SummarizerAttention._new_optimizer()
        summarizer.encoder.compile(optimizer=optimizer)
        summarizer.decoder.compile(optimizer=optimizer)
        summarizer.encoder.load_weights(encoder_path)
        summarizer.decoder.load_weights(decoder_path)
        summarizer.optimizer = summarizer.encoder.optimizer
        return summarizer

    @staticmethod
    def _new_optimizer() -> tf.keras.optimizers.Optimizer:
        return tf.keras.optimizers.Adam()



import json
import logging
from typing import Tuple, List

import tensorflow as tf
from headliner.model.summarizer_transformer import SummarizerTransformer

from headliner.model import SummarizerAttention
from sklearn.model_selection import train_test_split
from tensorflow_datasets.core.features.text import SubwordTextEncoder
from transformers import BertTokenizer

from headliner.model.summarizer_bert import SummarizerBert
from headliner.preprocessing import Preprocessor, Vectorizer
from headliner.trainer import Trainer


def read_data_json(file_path: str,
                   max_sequence_length: int) -> List[Tuple[str, str]]:
    with open(file_path, 'r', encoding='utf-8') as f:
        data_out = json.load(f)
        return [d for d in zip(data_out['desc'], data_out['heads']) if len(d[0].split(' ')) <= max_sequence_length]


def read_data(file_path: str) -> List[Tuple[str, str]]:
    data_out = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for l in f.readlines():
            x, y = l.strip().split('\t')
            data_out.append((x, y))
        return data_out


if __name__ == '__main__':

    train = read_data('/Users/cschaefe/datasets/en_ger.txt')[0:1000]

    train_data, val_data = train_test_split(train, test_size=10, shuffle=True, random_state=42)
    logging.getLogger("transformers.tokenization_utils").setLevel(logging.ERROR)
    tokenizer_input = BertTokenizer.from_pretrained('bert-base-uncased')
    tokenizer_target = BertTokenizer.from_pretrained('bert-base-german-cased')



    preprocessor = Preprocessor(start_token='[CLS]',
                                end_token='[SEP]',
                                filter_pattern=None,
                                add_input_start_end=False,
                                punctuation_pattern=None)
    train_prep = [preprocessor(t) for t in train_data]


    vectorizer = Vectorizer(tokenizer_input, tokenizer_target, max_input_len=512)

    p = preprocessor(train[0])
    v = vectorizer(p)
    print(p)


    summarizer = SummarizerBert(num_heads=8,
                                num_layers_encoder=0,
                                num_layers_decoder=1,
                                feed_forward_dim=1024,
                                embedding_size_encoder=768,
                                embedding_size_decoder=768,
                                bert_embedding_encoder='bert-base-uncased',
                                bert_embedding_decoder='bert-base-german-cased',
                                dropout_rate=0,
                                max_prediction_len=30)

    pred_beam = summarizer.predict_beam_search(train_data[0][0])
    for p in pred_beam:
        print(p)

    print('pred:')
    print(summarizer.predict(train_data[0][0]))
    print('target:')
    print(train_data[0][0])
    print(train_data[0][1])



import json
from typing import Tuple, List

from matplotlib import pyplot as plt
from tensorflow import squeeze

from headliner.model.summarizer_transformer import SummarizerTransformer


def plot_attention_weights(summarizer, pred_vectors, layer_name):
    fig = plt.figure(figsize=(16, 8))
    input_text, _ = pred_vectors['preprocessed_text']
    input_sequence = summarizer.vectorizer.encode_input(input_text)
    pred_sequence = pred_vectors['predicted_sequence']
    attention = squeeze(pred_vectors['attention_weights'][layer_name])
    for head in range(attention.shape[0]):
        ax = fig.add_subplot(1, 2, head + 1)
        ax.matshow(attention[head][:-1, :], cmap='viridis')
        fontdict = {'fontsize': 10}
        ax.set_xticks(range(len(input_sequence)))
        ax.set_yticks(range(len(pred_sequence)))
        ax.set_ylim(len(pred_sequence) - 1.5, -0.5)
        ax.set_xticklabels(
            [summarizer.vectorizer.decode_input([i]) for i in input_sequence],
            fontdict=fontdict,
            rotation=90)
        ax.set_yticklabels([summarizer.vectorizer.decode_output([i])
                               for i in pred_sequence ], fontdict=fontdict)
        ax.set_xlabel('Head {}'.format(head + 1))
    plt.tight_layout()
    plt.show()

def read_data_json(file_path: str,
                   max_sequence_length: int) -> List[Tuple[str, str]]:
    with open(file_path, 'r', encoding='utf-8') as f:
        data_out = json.load(f)
        return [d for d in zip(data_out['desc'], data_out['heads']) if len(d[0].split(' ')) <= max_sequence_length]


def read_data(file_path: str) -> List[Tuple[str, str]]:
    data_out = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for l in f.readlines():
            x, y = l.strip().split('\t')
            data_out.append((x, y))
        return data_out


if __name__ == '__main__':

    summarizer = SummarizerTransformer.load('/tmp/summarizer_20191010_170029')
    pred_vecs = summarizer.predict_vectors('This is old news.', '')
    sentence = pred_vecs['preprocessed_text'][0]
    attention = pred_vecs['attention_weights']
    print(attention)
    plot_attention_weights(summarizer, pred_vecs, 'decoder_layer1_block2')








import re
from typing import Tuple


class Preprocessor:

    def __init__(self,
                 start_token='<start>',
                 end_token='<end>',
                 punctuation_pattern='([!.?,])',
                 filter_pattern='(["#$%&()*+/:;<=>@[\\]^_`{|}~\t\n])',
                 add_input_start_end=True,
                 lower_case=True,
                 hash_numbers=True):
        """
        Initializes the preprocessor.

        Args:
            start_token: Unique start token to be inserted at the beginning of the target text.
            end_token: Unique end token to be attached at the end of a target text.
            punctuation_pattern: Regex pattern for punktuation that is splitted from the tokens.
            filter_pattern: Regex pattern for characters to be removed from the text.
            add_input_start_end: Whether to add start and end token to input sequence.
            lower_case: Whether to perform lower casing.
            hash_numbers: Whether to replace numbers by a #.
        """
        self.start_token = start_token
        self.end_token = end_token
        self.punctuation_pattern = punctuation_pattern
        self.filter_pattern = filter_pattern
        self.add_input_start_end = add_input_start_end
        self.lower_case = lower_case
        self.hash_numbers = hash_numbers

    def __call__(self, data: Tuple[str, str]) -> Tuple[str, str]:
        """ Performs regex logic for string cleansing and attaches start and end tokens to the text. """
        text_encoder, text_decoder = self.normalize_string(data[0]), self.normalize_string(data[1])
        if self.add_input_start_end:
            text_encoder = self.start_token + ' ' + text_encoder + ' ' + self.end_token
        text_decoder = self.start_token + ' ' + text_decoder + ' ' + self.end_token
        return text_encoder, text_decoder

    def normalize_string(self, s: str) -> str:
        if self.lower_case:
            s = s.lower()
        if self.filter_pattern is not None:
            s = re.sub(self.filter_pattern, '', s)
        if self.hash_numbers:
            s = re.sub(r'\d+', '#', s)
        if self.punctuation_pattern is not None:
            s = re.sub(self.punctuation_pattern, r' \1', s)
        s = re.sub(r'\s+', r' ', s)
        return s



from typing import Tuple, List

from headliner.preprocessing.tokenizer import Tokenizer


class Vectorizer:
    """
    Transforms tuples of text into tuples of vector sequences.
    """

    def __init__(self,
                 tokenizer_encoder: Tokenizer,
                 tokenizer_decoder: Tokenizer,
                 max_input_len=None,
                 max_output_len=None) -> None:
        """
        Initializes the vectorizer.

        Args:
            tokenizer_encoder: Tokenizer that encodes the input text.
            tokenizer_decoder: Tokenizer that encodes the target text.
            max_input_len (optional): Maximum length of input sequence,
                longer sequences will be truncated.
            max_output_len (optional): Maximum length of target sequence,
                longer sequences will be truncated and shorter sequences
                will be padded to max len.
        """
        self.encoding_dim = tokenizer_encoder.vocab_size + 1
        self.decoding_dim = tokenizer_decoder.vocab_size + 1
        self.max_input_len = max_input_len
        self.max_output_len = max_output_len
        self._tokenizer_encoder = tokenizer_encoder
        self._tokenizer_decoder = tokenizer_decoder

    def __call__(self, data: Tuple[str, str]) -> Tuple[List[int], List[int]]:
        """
        Encodes preprocessed strings into sequences of one-hot indices.
        """
        text_encoder, text_decoder = data[0], data[1]
        vec_encoder = self._tokenizer_encoder.encode(text_encoder)
        vec_decoder = self._tokenizer_decoder.encode(text_decoder)
        if self.max_input_len is not None:
            if len(vec_encoder) > self.max_input_len:
                vec_encoder = vec_encoder[:self.max_input_len-1] + [vec_encoder[-1]]
        if self.max_output_len is not None:
            if len(vec_decoder) > self.max_output_len:
                vec_decoder = vec_decoder[:self.max_output_len-1] + [vec_decoder[-1]]
            else:
                vec_decoder = vec_decoder + [0] * (self.max_output_len - len(vec_decoder))

        return vec_encoder, vec_decoder

    def encode_input(self, text: str) -> List[int]:
        return self._tokenizer_encoder.encode(text)

    def encode_output(self, text: str) -> List[int]:
        return self._tokenizer_decoder.encode(text)

    def decode_input(self, sequence: List[int]) -> str:
        return self._tokenizer_encoder.decode(sequence)

    def decode_output(self, sequence: List[int]) -> str:
        return self._tokenizer_decoder.decode(sequence)



import numpy as np
import tensorflow as tf


def get_angles(pos, i, embedding_size):
    angle_rates = 1 / np.power(10000, (2 * (i // 2)) / np.float32(embedding_size))
    return pos * angle_rates


def positional_encoding(max_len, embedding_size):
    angle_rads = get_angles(np.arange(max_len)[:, np.newaxis],
                            np.arange(embedding_size)[np.newaxis, :],
                            embedding_size)

    angle_rads[:, 0::2] = np.sin(angle_rads[:, 0::2])
    angle_rads[:, 1::2] = np.cos(angle_rads[:, 1::2])
    pos_encoding = angle_rads[np.newaxis, ...]
    return tf.cast(pos_encoding, dtype=tf.float32)


def create_padding_mask(seq):
    seq = tf.cast(tf.math.equal(seq, 0), tf.float32)
    return seq[:, tf.newaxis, tf.newaxis, :]


def create_look_ahead_mask(size):
    mask = 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0)
    return mask


def scaled_dot_product_attention(q, k, v, mask):
    matmul_qk = tf.matmul(q, k, transpose_b=True)
    dk = tf.cast(tf.shape(k)[-1], tf.float32)
    scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)
    if mask is not None:
        scaled_attention_logits += (mask * -1e9)
    attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)
    output = tf.matmul(attention_weights, v)

    return output, attention_weights


def point_wise_feed_forward_network(embedding_size, feed_forward_dim):
    return tf.keras.Sequential([
        tf.keras.layers.Dense(feed_forward_dim, activation='relu'),
        tf.keras.layers.Dense(embedding_size)
    ])


def create_masks(inp, tar):
    enc_padding_mask = create_padding_mask(inp)
    dec_padding_mask = create_padding_mask(inp)
    look_ahead_mask = create_look_ahead_mask(tf.shape(tar)[1])
    dec_target_padding_mask = create_padding_mask(tar)
    combined_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)
    return enc_padding_mask, combined_mask, dec_padding_mask


class MultiHeadAttention(tf.keras.layers.Layer):

    def __init__(self,
                 embedding_size: int,
                 num_heads: int) -> None:
        super(MultiHeadAttention, self).__init__()
        self.num_heads = num_heads
        self.embedding_size = embedding_size
        assert embedding_size % self.num_heads == 0
        self.depth = embedding_size // self.num_heads
        self.wq = tf.keras.layers.Dense(embedding_size)
        self.wk = tf.keras.layers.Dense(embedding_size)
        self.wv = tf.keras.layers.Dense(embedding_size)
        self.dense = tf.keras.layers.Dense(embedding_size)

    def split_heads(self, x, batch_size):
        x = tf.reshape(x, (batch_size, -1, self.num_heads, self.depth))
        return tf.transpose(x, perm=[0, 2, 1, 3])

    def call(self, v, k, q, mask):
        batch_size = tf.shape(q)[0]
        q = self.wq(q)
        k = self.wk(k)
        v = self.wv(v)
        q = self.split_heads(q, batch_size)
        k = self.split_heads(k, batch_size)
        v = self.split_heads(v, batch_size)
        scaled_attention, attention_weights = scaled_dot_product_attention(q, k, v, mask)
        scaled_attention = tf.transpose(scaled_attention, perm=[0, 2, 1, 3])
        concat_attention = tf.reshape(scaled_attention, (batch_size, -1, self.embedding_size))
        output = self.dense(concat_attention)
        return output, attention_weights


class EncoderLayer(tf.keras.layers.Layer):

    def __init__(self,
                 embedding_size: int,
                 num_heads: int,
                 feed_forward_dim: int,
                 dropout_rate=0.1) -> None:
        super(EncoderLayer, self).__init__()
        self.mha = MultiHeadAttention(embedding_size, num_heads)
        self.ffn = point_wise_feed_forward_network(embedding_size, feed_forward_dim)
        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = tf.keras.layers.Dropout(dropout_rate)
        self.dropout2 = tf.keras.layers.Dropout(dropout_rate)

    def call(self, x, training, mask):
        attn_output, _ = self.mha(x, x, x, mask)  # (batch_size, input_seq_len, embedding_size)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(x + attn_output)  # (batch_size, input_seq_len, embedding_size)
        ffn_output = self.ffn(out1)  # (batch_size, input_seq_len, embedding_size)
        ffn_output = self.dropout2(ffn_output, training=training)
        out2 = self.layernorm2(out1 + ffn_output)  # (batch_size, input_seq_len, embedding_size)
        return out2


class DecoderLayer(tf.keras.layers.Layer):

    def __init__(self,
                 embedding_size: int,
                 num_heads: int,
                 feed_forward_dim: int,
                 dropout_rate=0.1) -> None:
        super(DecoderLayer, self).__init__()
        self.mha1 = MultiHeadAttention(embedding_size, num_heads)
        self.mha2 = MultiHeadAttention(embedding_size, num_heads)
        self.ffn = point_wise_feed_forward_network(embedding_size, feed_forward_dim)
        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm3 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.dropout1 = tf.keras.layers.Dropout(dropout_rate)
        self.dropout2 = tf.keras.layers.Dropout(dropout_rate)
        self.dropout3 = tf.keras.layers.Dropout(dropout_rate)

    def call(self,
             x,
             enc_output,
             training,
             look_ahead_mask,
             padding_mask):
        attn1, attn_weights_block1 = self.mha1(x, x, x, look_ahead_mask)
        attn1 = self.dropout1(attn1, training=training)
        out1 = self.layernorm1(attn1 + x)
        attn2, attn_weights_block2 = self.mha2(enc_output, enc_output, out1, padding_mask)
        attn2 = self.dropout2(attn2, training=training)
        out2 = self.layernorm2(attn2 + out1)
        ffn_output = self.ffn(out2)
        ffn_output = self.dropout3(ffn_output, training=training)
        out3 = self.layernorm3(ffn_output + out2)

        return out3, attn_weights_block1, attn_weights_block2



from typing import Iterable, Callable

import tensorflow as tf


class DatasetGenerator:

    def __init__(self,
                 batch_size: int,
                 shuffle_buffer_size=None):
        self.batch_size = batch_size
        self.shuffle_buffer_size = shuffle_buffer_size

    def __call__(self, data_generator_func: Callable[..., Iterable]) -> tf.data.Dataset:
        """
        Initializes a dataset generator.

        Args:
            data_generator_func: Callable that returns an iterable over the data to be batched, e.g. lambda: [1, 2, 3].
        """
        tensor_types = (tf.int32, tf.int32)
        tensor_shapes = (tf.TensorShape([None]), tf.TensorShape([None]))
        dataset = tf.data.Dataset.from_generator(data_generator_func, tensor_types, tensor_shapes)
        if self.shuffle_buffer_size is not None:
            dataset = dataset.shuffle(self.shuffle_buffer_size)
        dataset = dataset.padded_batch(batch_size=self.batch_size,
                                       padded_shapes=([None], [None]),
                                       drop_remainder=True)
        return dataset



import unittest

from headliner.preprocessing.preprocessor import Preprocessor


class TestPreprocessor(unittest.TestCase):

    def test_preprocessing(self):
        preprocessor = Preprocessor()
        data = (('First text!', 'first head'), ('2-nd täxt', 'Second head'))
        data_preprocessed = [preprocessor(d) for d in data]
        self.assertEqual(('<start> first text ! <end>', '<start> first head <end>'), data_preprocessed[0])
        self.assertEqual(('<start> #-nd täxt <end>', '<start> second head <end>'), data_preprocessed[1])

        preprocessor = Preprocessor(start_token='<start>',
                                    end_token='<end>',
                                    lower_case=True,
                                    hash_numbers=False)
        data_preprocessed = [preprocessor(d) for d in data]
        self.assertEqual(('<start> 2-nd täxt <end>', '<start> second head <end>'), data_preprocessed[1])

        preprocessor = Preprocessor(start_token='<start>',
                                    end_token='<end>',
                                    lower_case=False,
                                    hash_numbers=True)
        data_preprocessed = [preprocessor(d) for d in data]
        self.assertEqual(('<start> #-nd täxt <end>', '<start> Second head <end>'), data_preprocessed[1])




import unittest

from headliner.evaluation.bleu_scorer import BleuScorer


class TestBleuScorer(unittest.TestCase):

    def test_score(self):
        bleu_scorer = BleuScorer(tokens_to_ignore={'<end>', '<unk>'})

        text_preprocessed = ('', 'this is a test')
        pred = 'this is a test'
        score = bleu_scorer({'preprocessed_text': text_preprocessed, 'predicted_text': pred})
        self.assertAlmostEqual(1, score, 5)

        text_preprocessed = ('', 'this is a test')
        pred = 'this is a test <end>'
        score = bleu_scorer({'preprocessed_text': text_preprocessed, 'predicted_text': pred})
        self.assertAlmostEqual(1, score, 5)

        text_preprocessed = ('', 'it is a guide to action which ensures that the military '
                                 'always obeys the commands of the party')
        pred = 'it is a guide to action that ensures that the military will ' \
               'forever <unk> <unk> <unk> heed Party commands <end>'
        score = bleu_scorer({'preprocessed_text': text_preprocessed, 'predicted_text': pred})
        self.assertAlmostEqual(0.4138, score, 3)



import tensorflow as tf

from headliner.model.summarizer import Summarizer


class ModelCheckpointCallback(tf.keras.callbacks.Callback):
    """
    Callback for checkpointing summarizer models.
    """

    def __init__(self,
                 file_path: str,
                 summarizer: Summarizer,
                 monitor='loss_val',
                 mode='min') -> None:

        """
        Initializes the Callback.

        Args:
            file_path: Path for saving the model (a directory). If existing, the model will be overwritten.
            summarizer: Summarizer to checkpoint.
            monitor: Name of the score monitor for improvements.
            mode: If set to 'min' a decrease of the monitored score is seen as an improvement, otherwise an increase.
        """

        super().__init__()
        self.file_path = file_path
        self.summarizer = summarizer
        self.monitor = monitor
        self.mode = mode
        self.best_score = None

    def on_epoch_end(self, batch, logs=None) -> None:
        if logs is None:
            logs = {}
        if self.file_path is None:
            return
        score = logs[self.monitor]
        score_is_better = False
        if self.best_score is None:
            score_is_better = True
        else:
            if self.mode == 'min' and score < self.best_score:
                score_is_better = True
            if self.mode == 'max' and score > self.best_score:
                score_is_better = True
        if score_is_better:
            print('{score_name} improved from {prev} to {current}, '
                  'saving summarizer to {path}'.format(score_name=self.monitor,
                                                       prev=self.best_score,
                                                       current=score,
                                                       path=self.file_path))
            self.best_score = score
            self.summarizer.save(self.file_path)
        else:
            print('{score_name} did not improve.'.format(score_name=self.monitor))



import itertools
import random
import unittest

from headliner.preprocessing.bucket_generator import BucketGenerator


class TestBucketGenerator(unittest.TestCase):

    def test_generate_batches_nonrandom(self):
        data = [[i] * i for i in range(10, 0, -1)]
        bucket_generator = BucketGenerator(lambda e: len(e),
                                           batch_size=2,
                                           buffer_size_batches=100,
                                           shuffle=False)
        buckets_gen = bucket_generator(data)
        result = [el for el in buckets_gen]
        expected = [[i] * i for i in range(1, 11)]
        self.assertEqual(expected, result)

    def test_generate_batches_random(self):
        data = [[i] * i for i in range(100, 0, -1)]
        random.shuffle(data)
        bucket_generator = BucketGenerator(lambda e: len(e),
                                           batch_size=2,
                                           buffer_size_batches=100,
                                           batches_to_bucket=10,
                                           shuffle=True,
                                           seed=42)
        buckets_gen = bucket_generator(data)
        result = [el for el in buckets_gen]

        # check whether all elements are returned
        expected_elements = list(itertools.chain.from_iterable(data))
        expected_elements.sort()
        result_elements = list(itertools.chain.from_iterable(result))
        result_elements.sort()
        self.assertEqual(expected_elements, result_elements)

        # check whether sequences of similar length are bucketed together
        # -> compare sum of length difference within batches against non-bucketed data
        raw_total_length_diff = 0
        result_total_length_diff = 0
        for i in range(0, len(data), 2):
            first_seq_raw, second_seq_raw = data[i:i + 2]
            first_seq_bucket, second_seq_bucket = result[i:i + 2]
            raw_total_length_diff += abs(len(second_seq_raw) - len(first_seq_raw))
            result_total_length_diff += abs(len(second_seq_bucket) - len(first_seq_bucket))
        self.assertTrue(result_total_length_diff < raw_total_length_diff / 4)



import abc
from typing import Dict, Union

import numpy as np


class Scorer(abc.ABC):

    def __call__(self, prediction: [Dict[str, Union[str, np.array]]]) -> float:
        """
        Evaluates prediction.

        Args:
            prediction: Dictionary providing all information about a model prediction such as
            output string, logits etc.

        Returns: Prediction score as float.
        """

        raise NotImplementedError()



import json
import logging
from typing import Tuple, List

from headliner.model.summarizer_transformer import SummarizerTransformer
from sklearn.model_selection import train_test_split
from tensorflow_datasets.core.features.text import SubwordTextEncoder

from headliner.model import SummarizerAttention
from headliner.preprocessing import Preprocessor, Vectorizer
from headliner.trainer import Trainer


def read_data_json(file_path: str,
                   max_sequence_length: int) -> List[Tuple[str, str]]:
    with open(file_path, 'r', encoding='utf-8') as f:
        data_out = json.load(f)
        return [d for d in zip(data_out['desc'], data_out['heads']) if len(d[0].split(' ')) <= max_sequence_length]


def read_data(file_path: str) -> List[Tuple[str, str]]:
    data_out = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for l in f.readlines():
            x, y = l.strip().split('\t')
            data_out.append((x, y))
        return data_out


if __name__ == '__main__':
    data_raw = read_data_json('/Users/cschaefe/datasets/welt_dedup.json', 500)

    preprocessor = Preprocessor()


    def filter_func(data):
        inp, out = data
        inp_tokens, out_tokens = set(inp.split()), set(out.split())
        ratio = len(out_tokens.intersection(inp_tokens)) / float(len(out_tokens))
        return ratio > 0.8
    data_filtered = [d for d in data_raw if filter_func(preprocessor(d))]

    print('len data filtered {}'.format(len(data_filtered)))

    train_data, val_data = train_test_split(data_raw, test_size=500, shuffle=True, random_state=42)
    logging.getLogger("transformers.tokenization_utils").setLevel(logging.ERROR)

    train_prep = [preprocessor(t) for t in train_data]
    inputs_prep = [t[0] for t in train_prep]
    targets_prep = [t[1] for t in train_prep]
    tokenizer_input = SubwordTextEncoder.build_from_corpus(
        inputs_prep, target_vocab_size=2 ** 13,
        reserved_tokens=[preprocessor.start_token, preprocessor.end_token])
    tokenizer_target = SubwordTextEncoder.build_from_corpus(
        targets_prep, target_vocab_size=2 ** 13,
        reserved_tokens=[preprocessor.start_token, preprocessor.end_token])
    vectorizer = Vectorizer(tokenizer_input, tokenizer_target)

    summarizer = SummarizerTransformer(embedding_size=64,
                                       max_prediction_len=40)
    summarizer.init_model(preprocessor, vectorizer)

    trainer = Trainer(steps_per_epoch=100,
                      batch_size=32,
                      steps_to_log=5,
                      tensorboard_dir='/tmp/tens_bert/attention_lstm_noprep',
                      max_output_len=40,
                      )

    trainer.train(summarizer, train_data, val_data=val_data)



from random import Random
from typing import Iterable, Callable, Iterator, List


class BucketGenerator:
    """
    Performs bucketing of elements in a dataset by length.
    """

    def __init__(self,
                 element_length_function: Callable[..., int],
                 batch_size: int,
                 buffer_size_batches=1000,
                 batches_to_bucket=10,
                 shuffle=True,
                 seed=None) -> None:
        """
        Initializes the BucketGenerator.

        Args:
            element_length_function: Element_length_function: function from element in the dataset to int that
            determines the length of the element.
            batch_size: The size of the batches to bucket the sequences into
                buffer_size_batches: buffer_size_batches: number of batches to keep in internal memory.
            batches_to_bucket: Number of batches in buffer to use for bucketing.
                If set to buffer_size_batches, the resulting batches will be deterministic.
            shuffle: Whether to shuffle elements across batches and the resulting buckets.
            seed: Seed for shuffling.
        """

        self.sequence_length_function = element_length_function
        self.batch_size = batch_size
        self.buffer_size_batches = buffer_size_batches
        self.batches_to_shuffle = batches_to_bucket
        self.shuffle = shuffle
        self.random = Random(seed)

    def __call__(self, data: Iterable) -> Iterable:
        """
        Returns iterable of data with elements ordered by bucketed sequence lengths, e.g for batch size = 2 the
        transformation could look like this:
        [1], [3, 3, 3], [1], [4, 4, 4, 4] -> [1], [1], [3, 3, 3], [4, 4, 4, 4]
        """
        data_iter = iter(data)
        bucket_gen = self._generate_buckets(data_iter)
        return bucket_gen

    def _generate_buckets(self, data_iter: Iterator) -> List[List]:
        buffered_data = self._fetch_buffered_data(data_iter)
        while len(buffered_data) > 0:
            buckets = self._to_buckets(buffered_data)
            for bucket in buckets:
                for element in bucket:
                    yield element
            buffered_data = self._fetch_buffered_data(data_iter)
            del buckets

    def _to_buckets(self, buffered_data: List) -> List[List]:
        self._shuffle_if_required(buffered_data)
        buffered_data = self._sort_blocks(buffered_data)
        buckets = []
        for i in range(0, len(buffered_data), self.batch_size):
            bucket = buffered_data[i:i + self.batch_size]
            if len(bucket) == self.batch_size:
                buckets.append(bucket)
        self._shuffle_if_required(buckets)
        return buckets

    def _sort_blocks(self, buffered_data: List) -> List:
        block_size = self.batches_to_shuffle * self.batch_size
        buffered_data_sorted = []
        for i in range(0, len(buffered_data), block_size):
            sorted_block = buffered_data[i:i + block_size]
            sorted_block.sort(key=self.sequence_length_function)
            buffered_data_sorted.extend(sorted_block)
        return buffered_data_sorted

    def _fetch_buffered_data(self, data_iter: Iterator):
        buffered_data = []
        for _ in range(self.buffer_size_batches * self.batch_size):
            try:
                buffered_data.append(next(data_iter))
            except StopIteration:
                pass
        return buffered_data

    def _shuffle_if_required(self, list_to_shuffle):
        if self.shuffle:
            self.random.shuffle(list_to_shuffle)



from typing import Callable

import tensorflow as tf

from headliner.model.summarizer import Summarizer


class ValidationCallback(tf.keras.callbacks.Callback):
    """
    Callback for validation loss.
    """

    def __init__(self,
                 summarizer: Summarizer,
                 val_dataset: tf.data.Dataset,
                 loss_function: Callable[[tf.Tensor, tf.Tensor], tf.Tensor],
                 batch_size: int) -> None:
        """
        Initializes the Callback.

        Args:
            summarizer: Summarizer to validate.
            val_dataset: Validation dataset to validate the model on.
            loss_function: Loss function to apply to calculate the validation score.
            batch_size: Batch size of the validation dataset, needed for initializing the model.
        """

        super().__init__()
        self.batch_size = batch_size
        self.summarizer = summarizer
        self.loss_function = loss_function
        self.val_dataset = val_dataset
        self.train_step = summarizer.new_train_step(self.loss_function,
                                                    self.batch_size,
                                                    apply_gradients=False)

    def on_epoch_end(self, batch, logs=None) -> None:
        if logs is None:
            logs = {}
        val_loss, count_batches_val = 0, 0
        for test_source_seq, test_target_seq in self.val_dataset.take(-1):
            val_loss_batch = self.train_step(test_source_seq, test_target_seq)
            val_loss += val_loss_batch
            count_batches_val += 1
        if count_batches_val == 0:
            raise ValueError('Tried to validate on empty validation dataset, possibly due to batch size '
                             'exceeding validation data size.')
        logs['loss_val'] = float(val_loss / count_batches_val)



import os
import shutil
import tempfile
import unittest

import numpy as np
import tensorflow as tf


from headliner.losses import masked_crossentropy
from headliner.model.summarizer_basic import SummarizerBasic
from headliner.preprocessing.keras_tokenizer import KerasTokenizer
from headliner.preprocessing.preprocessor import Preprocessor
from headliner.preprocessing.vectorizer import Vectorizer


class TestSummarizer(unittest.TestCase):

    def setUp(self) -> None:
        np.random.seed(42)
        tf.random.set_seed(42)
        self.temp_dir = tempfile.mkdtemp(prefix='TestSummarizerTmp')

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def test_serde_happy_path(self) -> None:
        preprocessor = Preprocessor()
        tokenizer = KerasTokenizer(oov_token='<unk>')
        tokenizer.fit(['a b c {} {}'.format(
            preprocessor.start_token, preprocessor.end_token)])
        vectorizer = Vectorizer(tokenizer, tokenizer)
        summarizer = SummarizerBasic(lstm_size=10,
                                     max_prediction_len=10,
                                     embedding_decoder_trainable=False,
                                     embedding_size=10)
        summarizer.init_model(preprocessor=preprocessor,
                              vectorizer=vectorizer)

        # we need at least a train step to init the weights
        train_step = summarizer.new_train_step(masked_crossentropy, batch_size=1, apply_gradients=True)
        train_seq = tf.convert_to_tensor(np.array([[1, 1, 1]]), dtype=tf.int32)
        train_step(train_seq, train_seq)

        save_dir = os.path.join(self.temp_dir, 'summarizer_serde_happy_path')
        summarizer.save(save_dir)
        summarizer_loaded = SummarizerBasic.load(save_dir)
        self.assertEqual(10, summarizer_loaded.lstm_size)
        self.assertEqual(10, summarizer_loaded.max_prediction_len)
        self.assertIsNotNone(summarizer_loaded.preprocessor)
        self.assertIsNotNone(summarizer_loaded.vectorizer)
        self.assertIsNotNone(summarizer_loaded.encoder)
        self.assertIsNotNone(summarizer_loaded.decoder)
        self.assertTrue(summarizer_loaded.encoder.embedding.trainable)
        self.assertFalse(summarizer_loaded.decoder.embedding.trainable)
        self.assertIsNotNone(summarizer_loaded.optimizer)

        pred = summarizer.predict_vectors('a c', '')
        pred_loaded = summarizer_loaded.predict_vectors('a c', '')
        np.testing.assert_almost_equal(
            pred['logits'], pred_loaded['logits'], decimal=6)


from sklearn.model_selection import train_test_split

from headliner.training_runner import read_data_json
from headliner.model.summarizer_bert import SummarizerBert

if __name__ == '__main__':

    train = read_data_json('/Users/cschaefe/datasets/welt_dedup.json', 400)[:1000]
    train_data, val_data = train_test_split(train, test_size=100, shuffle=True, random_state=32)
    path_to_model = '/Users/cschaefe/transformer_check_7'
    summarizer = SummarizerBert.load(path_to_model)
    summarizer.max_prediction_len = 30
    for val_index in range(100):
        text = val_data[val_index][0]
        target = val_data[val_index][1]
        print('(target) {}'.format(target))
        pred = summarizer.predict_beam_search(text, beam_width=15)
        for p in pred:
            print('{} {}'.format(p[0], p[1]))



import os
import pickle
from typing import Callable
from typing import Dict, Union

import numpy as np
import tensorflow as tf

from headliner.model.model_transformer import Transformer, create_masks
from headliner.model.summarizer import Summarizer
from headliner.preprocessing.preprocessor import Preprocessor
from headliner.preprocessing.vectorizer import Vectorizer


class SummarizerTransformer(Summarizer):

    def __init__(self,
                 max_prediction_len=20,
                 num_layers=1,
                 num_heads=2,
                 feed_forward_dim=512,
                 dropout_rate=0,
                 embedding_size=128,
                 embedding_encoder_trainable=True,
                 embedding_decoder_trainable=True,
                 max_sequence_len=10000):

        super().__init__()
        self.max_prediction_len = max_prediction_len
        self.embedding_size = embedding_size
        self.num_layers = num_layers
        self.num_heads = num_heads
        self.dropout_rate = dropout_rate
        self.feed_forward_dim = feed_forward_dim
        self.embedding_encoder_trainable = embedding_encoder_trainable
        self.embedding_decoder_trainable = embedding_decoder_trainable
        self.optimizer = SummarizerTransformer.new_optimizer()
        self.transformer = None
        self.embedding_shape_in = None
        self.embedding_shape_out = None
        self.max_sequence_len = max_sequence_len

    def __getstate__(self):
        """ Prevents pickle from serializing the transformer and optimizer """
        state = self.__dict__.copy()
        del state['transformer']
        del state['optimizer']
        return state

    def init_model(self,
                   preprocessor: Preprocessor,
                   vectorizer: Vectorizer,
                   embedding_weights_encoder=None,
                   embedding_weights_decoder=None) -> None:
        self.preprocessor = preprocessor
        self.vectorizer = vectorizer
        self.embedding_shape_in = (self.vectorizer.encoding_dim, self.embedding_size)
        self.embedding_shape_out = (self.vectorizer.decoding_dim, self.embedding_size)
        self.transformer = Transformer(num_layers=self.num_layers,
                                       num_heads=self.num_heads,
                                       feed_forward_dim=self.feed_forward_dim,
                                       embedding_shape_encoder=self.embedding_shape_in,
                                       embedding_shape_decoder=self.embedding_shape_out,
                                       embedding_encoder_trainable=self.embedding_encoder_trainable,
                                       embedding_decoder_trainable=self.embedding_decoder_trainable,
                                       embedding_weights_encoder=embedding_weights_encoder,
                                       embedding_weights_decoder=embedding_weights_decoder,
                                       dropout_rate=self.dropout_rate,
                                       max_sequence_len=self.max_sequence_len)
        self.transformer.compile(optimizer=self.optimizer)

    def new_train_step(self,
                       loss_function: Callable[[tf.Tensor], tf.Tensor],
                       batch_size: int,
                       apply_gradients=True):

        transformer = self.transformer
        optimizer = self.optimizer

        train_step_signature = [
            tf.TensorSpec(shape=(batch_size, None), dtype=tf.int32),
            tf.TensorSpec(shape=(batch_size, None), dtype=tf.int32),
        ]

        @tf.function(input_signature=train_step_signature)
        def train_step(inp, tar):
            tar_inp = tar[:, :-1]
            tar_real = tar[:, 1:]
            enc_padding_mask, combined_mask, dec_padding_mask = create_masks(inp, tar_inp)
            with tf.GradientTape() as tape:
                predictions, _ = transformer(inp, tar_inp,
                                             True,
                                             enc_padding_mask,
                                             combined_mask,
                                             dec_padding_mask)
                loss = loss_function(tar_real, predictions)
            if apply_gradients:
                gradients = tape.gradient(loss, transformer.trainable_variables)
                optimizer.apply_gradients(zip(gradients, transformer.trainable_variables))
            return loss

        return train_step

    def predict(self, text: str) -> str:
        return self.predict_vectors(text, '')['predicted_text']

    def predict_vectors(self, input_text: str, target_text: str) -> Dict[str, Union[str, np.array]]:
        text_preprocessed = self.preprocessor((input_text, target_text))
        en_inputs, _ = self.vectorizer(text_preprocessed)
        en_inputs = tf.expand_dims(en_inputs, 0)
        start_end_seq = self.vectorizer.encode_output(
            ' '.join([self.preprocessor.start_token, self.preprocessor.end_token]))
        de_start_index, de_end_index = start_end_seq[:1], start_end_seq[-1:]
        decoder_output = tf.expand_dims(de_start_index, 0)
        output = {'preprocessed_text': text_preprocessed,
                  'logits': [],
                  'attention_weights': [],
                  'predicted_sequence': []}
        for _ in range(self.max_prediction_len):
            enc_padding_mask, combined_mask, dec_padding_mask = create_masks(
                en_inputs, decoder_output)
            predictions, attention_weights = self.transformer(en_inputs,
                                                              decoder_output,
                                                              False,
                                                              enc_padding_mask,
                                                              combined_mask,
                                                              dec_padding_mask)

            predictions = predictions[:, -1:, :]
            pred_token_index = tf.cast(tf.argmax(predictions, axis=-1), tf.int32)
            decoder_output = tf.concat([decoder_output, pred_token_index], axis=-1)
            if pred_token_index != 0:
                output['logits'].append(np.squeeze(predictions.numpy()))
                output['attention_weights'] = attention_weights
                output['predicted_sequence'].append(int(pred_token_index))
                if pred_token_index == de_end_index:
                    break
        output['predicted_text'] = self.vectorizer.decode_output(output['predicted_sequence'])
        return output

    def save(self, out_path: str) -> None:
        if not os.path.exists(out_path):
            os.mkdir(out_path)
        summarizer_path = os.path.join(out_path, 'summarizer.pkl')
        transformer_path = os.path.join(out_path, 'transformer')
        with open(summarizer_path, 'wb+') as handle:
            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)
        self.transformer.save_weights(transformer_path, save_format='tf')

    @staticmethod
    def load(in_path: str):
        summarizer_path = os.path.join(in_path, 'summarizer.pkl')
        transformer_path = os.path.join(in_path, 'transformer')
        with open(summarizer_path, 'rb') as handle:
            summarizer = pickle.load(handle)
        summarizer.transformer = Transformer(num_layers=summarizer.num_layers,
                                             num_heads=summarizer.num_heads,
                                             feed_forward_dim=summarizer.feed_forward_dim,
                                             embedding_shape_encoder=summarizer.embedding_shape_in,
                                             embedding_shape_decoder=summarizer.embedding_shape_out,
                                             embedding_encoder_trainable=summarizer.embedding_encoder_trainable,
                                             embedding_decoder_trainable=summarizer.embedding_decoder_trainable,
                                             dropout_rate=summarizer.dropout_rate)
        optimizer = SummarizerTransformer.new_optimizer()
        summarizer.transformer.compile(optimizer=optimizer)
        summarizer.transformer.load_weights(transformer_path)
        summarizer.optimizer = summarizer.transformer.optimizer
        return summarizer

    @staticmethod
    def new_optimizer() -> tf.keras.optimizers.Optimizer:
        return tf.keras.optimizers.Adam()



import json
import logging
from typing import Tuple, List

import tensorflow as tf
from headliner.model.summarizer_transformer import SummarizerTransformer

from headliner.model import SummarizerAttention
from sklearn.model_selection import train_test_split
from tensorflow_datasets.core.features.text import SubwordTextEncoder
from transformers import BertTokenizer

from headliner.model.summarizer_bert import SummarizerBert
from headliner.preprocessing import Preprocessor, Vectorizer
from headliner.trainer import Trainer


def read_data_json(file_path: str,
                   max_sequence_length: int) -> List[Tuple[str, str]]:
    with open(file_path, 'r', encoding='utf-8') as f:
        data_out = json.load(f)
        return [d for d in zip(data_out['desc'], data_out['heads']) if len(d[0].split(' ')) <= max_sequence_length]


def read_data(file_path: str) -> List[Tuple[str, str]]:
    data_out = []
    with open(file_path, 'r', encoding='utf-8') as f:
        for l in f.readlines():
            x, y = l.strip().split('\t')
            data_out.append((x, y))
        return data_out


if __name__ == '__main__':
    from headliner.trainer import Trainer
    from headliner.model.summarizer_transformer import SummarizerTransformer

    train = read_data('/Users/cschaefe/datasets/en_ger.txt')[0:10000]

    train_data, val_data = train_test_split(train, test_size=100, shuffle=True, random_state=42)

    summarizer = SummarizerTransformer(embedding_size=64, max_prediction_len=20)
    trainer = Trainer(batch_size=1, steps_per_epoch=100, model_save_path='/tmp/summarizer')
    trainer.train(summarizer, train_data, val_data=val_data, num_epochs=2)

    pred_beam = summarizer.predict_beam_search(train_data[0][0])
    for p in pred_beam:
        print(p)

    print('pred:')
    print(summarizer.predict(train_data[0][0]))
    print('target:')
    print(train_data[0][0])
    print(train_data[0][1])
